#!/usr/bin/env python

from array import array
from itertools import combinations
import PIL.Image
import sys

def scale(v):
    return (v + 43) / 85

def point(rgb):
    return tuple(map(scale, rgb))

def index(p):
    return p[0] * 16 + p[1] * 4 + p[2]

def colour(i):
    return (255 * (i % 2), 255 * ((i / 2) % 2), 255 * ((i / 4) % 2))

def add(d, v):
    d[v] = (d.has_key(v) and d[v] or 0) + 1

def by_frequency(d):
    l = [(f, t) for (t, f) in d.items()]
    l.sort(reverse=True)
    return [i[1] for i in l]

def match(b, bases):
    return b in bases and b

def fallback(bases):
    for b in by_frequency(bases):
        if b not in ["_", "W"]:
            return b
    return by_frequency(bases)[0]

tones = [
    "___", "_BB", "_BB", "BBB", # 00x
    "_GG", "__C", "_BC", "BCC", # 01x
    "_GG", "GGC", "BCC", "CCC", # 02x
    "GGG", "GCC", "CCC", "CCC", # 03x
    "_RR", "_MM", "MMB", "MBB", # 10x
    "_YY", "_**", "_*B", "BBW", # 11x
    "_GY", "GGC", "*CC", "CCW", # 12x
    "GGY", "GGG", "GCC", "CCW", # 13x
    "RRR", "RRM", "RMM", "MMM", # 20x
    "RYY", "RRW", "RMW", "MMW", # 21x
    "YYY", "YYW", "**W", "WWW", # 22x
    "YYY", "YYW", "YWW", "WWW", # 23x
    "RRR", "RMM", "RMM", "MMW", # 30x
    "RRY", "RRY", "RMW", "MMW", # 31x
    "YYY", "YYW", "YYW", "WWW", # 32x
    "YYY", "YYW", "YYW", "WWW", # 33x
    ]

colours = ["_", "R", "G", "Y", "B", "M", "C", "W"]

if __name__ == "__main__":
    width = 320
    input_filename, output_filename = sys.argv[1:3]
    rotate = "-r" in sys.argv[3:]

    im = PIL.Image.open(input_filename)
    if rotate:
        im = im.rotate(270)
    w, h = im.size
    height = (width * h) / w
    im = im.resize((width, height))

    usage = []
    base_usage = []
    toned = []

    for row in range(0, height):
        u = {}
        usage.append(u)
        bu = {}
        base_usage.append(bu)
        tr = []
        toned.append(tr)
        for column in range(0, width):
            rgb = im.getpixel((column, row))
            p = point(rgb)
            i = index(p)
            t = tones[i]
            add(u, t)
            if t[0] != "*":
                add(bu, t[0])
            if t[1] != "*":
                add(bu, t[1])
            if t[2] != "*":
                add(bu, t[2])
            tr.append(t)

    chosen = []

    for row, (u, bu) in enumerate(zip(usage, base_usage)):
        light = row % 2
        best = 0
        best_bases = None
        best_missing = None
        best_map = None

        for bases in combinations(bu, min(len(bu), 4)):
            bases = dict([(base, bu[base]) for base in bases])
            count = 0
            missing = []
            tone_map = {}
            for tone, freq in u.items():
                base = match(tone[1], bases)
                if base:
                    tone_map[tone] = base
                    count += freq
                else:
                    base = match(light and tone[2] or tone[0], bases)
                    if base:
                        tone_map[tone] = base
                        count += freq / 2
                    else:
                        missing.append(tone)
            if count > best:
                best_bases = bases
                best_missing = missing
                best_map = tone_map
                best = count

        chosen.append((best, best_bases or bases, best_map or tone_map, best_missing or missing))

    output = []

    for row, (tr, ch) in enumerate(zip(toned, chosen)):
        o = []
        for column, t in enumerate(tr):
            best, bases, tone_map, missing = ch
            base = tone_map.get(t) or fallback(bases)
            o.append(base)
            i = colours.index(base)
            im.putpixel((column, row), colour(i))

        output.append("".join(o))

    im.save(output_filename)

# vim: tabstop=4 expandtab shiftwidth=4
